"""Hydrological analysis library built with xarray."""

from . import cc, indicators, utils

# Import top-level functions
# TODO: Decide which functions to import at the top level

__author__ = """Thomas-Charles Fortier Filion"""
__email__ = "tcff_hydro@outlook.com"
__version__ = "0.3.4"
